#!/usr/bin/env python
#
# Copyright 2010 Facebook
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#	 http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Python client library for the Facebook Platform.

This client library is designed to support the Graph API and the official
Facebook JavaScript SDK, which is the canonical way to implement
Facebook authentication. Read more about the Graph API at
http://developers.facebook.com/docs/api. You can download the Facebook
JavaScript SDK at http://github.com/facebook/connect-js/.
"""

import cgi
import hashlib
import time
import urllib
import pickle

# Find a JSON parser
try:
	import json
	_parse_json = lambda s: json.loads(s)
except ImportError:
	try:
		import simplejson
		_parse_json = lambda s: simplejson.loads(s)
	except ImportError:
		# For Google AppEngine
		from django.utils import simplejson
		_parse_json = lambda s: simplejson.loads(s)

class GraphAPI(object):
	def __init__(self, access_token = None, cache = None):
		self.access_token = access_token
		self.cache = None
		if cache != None:
			self.cacheFile = cache
			try:
				file = open(cache, "rb")
				self.cache = pickle.load(file)
			except IOError:
				self.cache = {}
	
	def request(self, path, args=None, post_args=None):
		"""Fetches the given path in the Graph API.

		We translate args to a valid query string. If post_args is given,
		we send a POST request to the given path with the given arguments.
		"""
		if args == None:
			args = ""
		key = path + args
		if self.access_token:
			if post_args is not None:
				post_args["access_token"] = self.access_token
			else:
				if len(args) > 0:
					args += "&"
				args += "access_token=" + self.access_token
		post_data = None if post_args is None else urllib.urlencode(post_args)
		file = urllib.urlopen("https://graph.facebook.com/" + path + "?" + args, post_data)
		try:
			response = _parse_json(file.read())
			if self.cache != None:
				self.cache[key] = response
		finally:
			file.close()
		return response
		
	def plainRequest(self, url):
		"""Use this to request an absolute URL that already has an access token in it."""
		file = urllib.urlopen(url, None)
		try:
			response = _parse_json(file.read())
			if self.cache != None:
				self.cache[url] = response
		finally:
			file.close()
		return response
	
	def plainRawRequest(self, url):
		"""Use this to request an absolute URL and get the raw byte data."""
		file = urllib.urlopen(url, None)
		try:
			response = file.read()
			if self.cache != None:
				self.cache[url] = response
		finally:
			file.close()
		return response
	
	def rawRequest(self, path, args=None, post_args=None):
		"""Use this to request a relative URL and get the raw byte data."""
		if args == None:
			args = ""
		key = path + args
		if self.access_token:
			if post_args is not None:
				post_args["access_token"] = self.access_token
			else:
				if len(args) > 0:
					args += "&"
				args += "access_token=" + self.access_token
		post_data = None if post_args is None else urllib.urlencode(post_args)
		file = urllib.urlopen("https://graph.facebook.com/" + path + "?" + args, post_data)
		data = file.read()
		if self.cache != None:
			self.cache[key] = data
		file.close()
		return data
	
	def saveCache(self):
		pickle.dump(self.cache, open(self.cacheFile, "wb"))

class CachedGraphAPI(object):
	def __init__(self, file):
		self.cache = pickle.load(open(file, "rb"))
	
	def request(self, path, args=None, post_args=None):
		"""Fetches the given path in the Graph API.

		We translate args to a valid query string. If post_args is given,
		we send a POST request to the given path with the given arguments.
		"""
		if args == None:
			args = ""
		key = path + args
		if key in self.cache:
			return self.cache[key]
		else:
			return []
		
	def plainRequest(self, url):
		"""Use this to request an absolute URL that already has an access token in it."""
		if url in self.cache:
			return self.cache[url]
		else:
			return []
	
	def plainRawRequest(self, url):
		"""Use this to request an absolute URL and get the raw byte data."""
		if url in self.cache:
			return self.cache[url]
		else:
			return ""
	
	def rawRequest(self, path, args=None, post_args=None):
		"""Use this to request a relative URL and get the raw byte data."""
		if args == None:
			args = ""
		key = path + args
		if key in self.cache:
			return self.cache[key]
		else:
			return ""
	
	def saveCache(self):
		pass

def get_user_from_cookie(cookies, app_id, app_secret):
	"""Parses the cookie set by the official Facebook JavaScript SDK.

	cookies should be a dictionary-like object mapping cookie names to
	cookie values.

	If the user is logged in via Facebook, we return a dictionary with the
	keys "uid" and "access_token". The former is the user's Facebook ID,
	and the latter can be used to make authenticated requests to the Graph API.
	If the user is not logged in, we return None.

	Download the official Facebook JavaScript SDK at
	http://github.com/facebook/connect-js/. Read more about Facebook
	authentication at http://developers.facebook.com/docs/authentication/.
	"""
	cookie = cookies.get("fbs_" + app_id, "")
	if not cookie: return None
	args = dict((k, v[-1]) for k, v in cgi.parse_qs(cookie.strip('"')).items())
	payload = "".join(k + "=" + args[k] for k in sorted(args.keys())
					  if k != "sig")
	sig = hashlib.md5(payload + app_secret).hexdigest()
	expires = int(args["expires"])
	if sig == args.get("sig") and (expires == 0 or time.time() < expires):
		return args
	else:
		return None