from pandac.PandaModules import *
import math
from direct.showbase.DirectObject import DirectObject
from uuid import uuid4
from random import random
import re
import htmlentitydefs

boldStartRegex = re.compile(r"<\s*(?:b|strong)\s*>", re.IGNORECASE)
boldEndRegex = re.compile(r"<\s*/\s*(?:b|strong)\s*>", re.IGNORECASE)
italicStartRegex = re.compile(r"<\s*(?:i|em)\s*>", re.IGNORECASE)
italicEndRegex = re.compile(r"<\s*/\s*(?:i|em)\s*>", re.IGNORECASE)
subscriptStartRegex = re.compile(r"<\s*sub\s*>", re.IGNORECASE)
subscriptEndRegex = re.compile(r"<\s*/\s*sub\s*>", re.IGNORECASE)
superscriptStartRegex = re.compile(r"<\s*sup\s*>", re.IGNORECASE)
superscriptEndRegex = re.compile(r"<\s*/\s*sup\s*>", re.IGNORECASE)
underlineStartRegex = re.compile(r"<\s*u\s*>", re.IGNORECASE)
underlineEndRegex = re.compile(r"<\s*/\s*u\s*>", re.IGNORECASE)
allTagsRegex = re.compile(r"<[^<]*?>")
unescapeRegex = re.compile(r"&#?\w+;")

def unescape(text):
    def fixup(m):
        text = m.group(0)
        if text[:2] == "&#":
            # character reference
            try:
                if text[:3] == "&#x":
                    return unichr(int(text[3:-1], 16))
                else:
                    return unichr(int(text[2:-1]))
            except ValueError:
                pass
        else:
            # named entity
            try:
                text = unichr(htmlentitydefs.name2codepoint[text[1:-1]])
            except KeyError:
                pass
        return text # leave as is
    return unescapeRegex.sub(fixup, text)

clock = None # Must be given by main

class Connection:
	def __init__(self, id, name):
		self.id = id
		self.name = name

class Page:
	def __init__(self, url):
		self.url = url

class MeshGraph(DirectObject):
	def __init__(self, source):
		self.drawer = MeshDrawer()
		self.drawer.setBudget(10000)
		self.node = self.drawer.getRoot()
		self.node.reparentTo(render)
		self.node.setDepthWrite(False)
		self.node.setTransparency(True)
		self.node.setTwoSided(False)
		self.node.setTexture(loader.loadTexture("circle.png"))
		self.node.setBin("fixed", 100)
		self.node.setLightOff(True)
		self.node.setShaderOff()
		self.node.node().setBounds(BoundingSphere((0, 0, 0), 100000)) 
		self.node.node().setFinal(True)
		self.segments = []
		self.nodeTable = dict()
		self.nodes = []
		self.source = source
		
		self.graphRoot = render.attachNewNode("graphRoot")
		
		self.mouseDownTime = 0.0
		self.enableRotation = False
		self.lastMousePosition = Point2()
		self.accept("mouse1", self.mouseHandler)
		self.accept("mouse1-up", self.mouseUpHandler)
		
		self.lastCameraFocusNode = None
		self.cameraFocusNode = None
		self.cameraFocusInterval = 0
		self.cameraPitch = math.pi * 0.5
		self.cameraYaw = 0.0
		
		self.lastHighlightedNode = None
		
		if base.cTrav == 0:
			self.traverser = CollisionTraverser("collision_traverser")
			base.cTrav = self.traverser
		else:
			self.traverser = base.cTrav
			self.traverser.clearColliders()
		
		self.world = OdeWorld()
		# Create a space and add a contactgroup to it to add the contact joints
		self.space = OdeHashSpace()
		self.space.setAutoCollideWorld(self.world)
		self.contactGroup = OdeJointGroup()
		self.space.setAutoCollideJointGroup(self.contactGroup)
		
		self.world.initSurfaceTable(1)
		self.world.setSurfaceEntry(0, 0, 1.0, 0.3, 7, 0.9, 0.00001, 0.0, 0.01)
		
		self.myId = None
		
		self.reservedKeywords = ["id", "name", "metadata", "paging", "data", "actions", "count", "offset", "length", "icon", "link", "source", "images", "width", "height", "created_time", "updated_time", "position"]
		
	def constructNodes(self, parent, radius, text, data, instance = None):
		if isinstance(data, dict):
			alreadyAddedChildren = [] if instance == None else instance.links.keys()
			children = [(key, value) for (key, value) in data.items() if key not in self.reservedKeywords and key not in alreadyAddedChildren]
			if "id" in data:
				id = str(data["id"])
				if instance == None:
					node = self.getNode(id)
					if node != None:
						if parent != None:
							parent.link(text, node)
							self.segments.append((parent, node))
						return node
				color = Vec4(0, 0.8, 1.0, 1)
				if "metadata" in data and "connections" in data["metadata"]:
					children.extend([(key, Connection(id, key)) for (key, value) in data["metadata"]["connections"].items()])
			else:
				id = None
				color = Vec4(0.2, 1, 0.2, 1)
			
			if "name" in data and data["name"] != None:
				label = text + ":\n" + data["name"]
			else:
				label = text
			
			if "data" in data:
				i = 1
				for item in data["data"]:
					children.append((str(i), item))
					i += 1
			
			if "paging" in data and len(children) > 0:
				children.append(("More...", Page(data["paging"]["next"])))
		elif isinstance(data, list):
			color = Vec4(1, 0.1, 0.1, 1)
			children = []
			i = 1
			for item in data:
				children.append((str(i), item))
				i += 1
			id = None
			label = text
		elif isinstance(data, Connection):
			color = Vec4(1, 0.5, 1, 1) if text == "picture" else Vec4(1, 1, 0, 1)
			children = []
			id = None
			label = text
		elif isinstance(data, Page):
			color = Vec4(0.2, 1, 0.2, 1)
			children = []
			id = None
			label = text
		elif text == "picture":
			color = Vec4(1, 0.5, 1, 1)
			children = []
			id = None
			label = text
		else:
			color = Vec4(1, 1, 1, 1)
			children = []
			id = None
			stringData = None
			if isinstance(data, unicode):
				stringData = data.encode("ascii", "replace")
			else:
				stringData = str(data)
			label = text + ":\n" + stringData
		
		if instance == None:
			instance = self.addNode(parent, radius, id, text, label, data)
			instance.node.setColor(color)
		
		if len(children) > 0:
			if parent == None:
				angleInterval = (2.0 * math.pi) / len(children)
			else:
				angleInterval = (1.0 * math.pi) / len(children)
			instance.childOrbitRadius = 10 + (5 if len(children) == 1 else (5 / math.sin(angleInterval)))
			for key, child in children:
				self.constructNodes(instance, instance.childOrbitRadius, key, child)
		return instance
	
	def addNode(self, parent, radius, id, key, text, data):
		n = MeshGraphNode(self.graphRoot, parent, radius, id, text, data, self.world, self.space)
		if parent != None:
			parent.link(key, n)
			self.segments.append((parent, n))
		self.nodeTable[n.id] = n
		self.nodes.append(n)
		if self.cameraFocusNode == None:
			self.cameraFocusNode = n
		if self.myId == None:
			self.myId = n.id
		return n
	
	def getNode(self, id):
		if id in self.nodeTable:
			return self.nodeTable[id]
		else:
			return None
	
	def mouseHandler(self):
		self.enableRotation = True
		self.lastMousePosition = base.mouseWatcherNode.getMouse()
		self.mouseDownTime = clock.time
	
	def mouseUpHandler(self):
		self.enableRotation = False
		
		if clock.time - self.mouseDownTime > 0.25:
			return
		
		def getCollisions(rayNP):
			queue = CollisionHandlerQueue()
			self.traverser.addCollider(rayNP, queue)
			self.traverser.traverse(self.graphRoot)
			self.traverser.clearColliders()
			queue.sortEntries()
			return queue
	
		if base.mouseWatcherNode.hasMouse():
			pickerNode = CollisionNode("mouseRay")
			pickerNP = camera.attachNewNode(pickerNode)
			pickerNode.setIntoCollideMask(BitMask32(0))
			pickerNode.setFromCollideMask(BitMask32(1))
			pickerRay = CollisionRay()
			mpos = base.mouseWatcherNode.getMouse()
			pickerRay.setFromLens(base.camNode, mpos.getX(), mpos.getY())
			pickerNode.addSolid(pickerRay)
			
			queue = getCollisions(pickerNP)
			for i in range(queue.getNumEntries()):
				hit = queue.getEntry(i)
				id = hit.getIntoNodePath().getParent().getName()
				if len(id) > 0:
					node = self.getNode(id)
					if node == None:
						continue
					if self.cameraFocusNode != None:
						self.cameraFocusNode.recalculateOffset(selected = False)
					self.lastCameraFocusNode = self.cameraFocusNode
					self.cameraFocusNode = node
					self.cameraFocusInterval = 0
					node.recalculateOffset(selected = True)
					if not node.expanded:
						node.expanded = True
						if isinstance(node.data, Connection):
							if node.data.name == "picture":
								node.setImage(self.source.rawRequest(node.data.id + "/" + node.data.name))
							elif node.data.name == "mutualfriends":
								self.constructNodes(node.parent, 0, "", self.source.request(node.data.id + "/" + node.data.name, "limit=5&user=" + self.myId), node)
							else:
								self.constructNodes(node.parent, 0, "", self.source.request(node.data.id + "/" + node.data.name, "limit=5"), node)
						elif isinstance(node.data, Page):
							self.constructNodes(node.parent, 0, "", self.source.plainRequest(node.data.url), node)
						elif isinstance(node.data, dict) and "id" in node.data:
							self.constructNodes(node.parent, 0, "", self.source.request(id, "metadata=1"), node)
						elif node.text == "picture":
							node.setImage(self.source.plainRawRequest(node.data))
					break
	
	def goBack(self):
		node = self.cameraFocusNode.parent
		if node == None:
			return
		
		if self.cameraFocusNode != None:
			self.cameraFocusNode.recalculateOffset(selected = False)
		self.lastCameraFocusNode = self.cameraFocusNode
		self.cameraFocusNode = node
		self.cameraFocusInterval = 0
		node.recalculateOffset(selected = True)
		
	def update(self, dt):
		if base.mouseWatcherNode.hasMouse():
			if self.lastHighlightedNode != None:
				self.lastHighlightedNode.node.setScale(self.lastHighlightedNode.defaultScale)
				self.lastHighlightedNode = None
			if self.enableRotation:
				mousePos = base.mouseWatcherNode.getMouse()
				self.cameraYaw -= mousePos.getX() - self.lastMousePosition.getX()
				self.cameraPitch = max(0.05, min(math.pi - 0.05, self.cameraPitch + (mousePos.getY() - self.lastMousePosition.getY())))
				self.lastMousePosition = Point2(mousePos)
			else:
				pickerNode = CollisionNode("mouseRay")
				pickerNP = camera.attachNewNode(pickerNode)
				pickerNode.setIntoCollideMask(BitMask32(0))
				pickerNode.setFromCollideMask(BitMask32(1))
				pickerRay = CollisionRay()
				mpos = base.mouseWatcherNode.getMouse()
				pickerRay.setFromLens(base.camNode, mpos.getX(), mpos.getY())
				pickerNode.addSolid(pickerRay)
				
				def getCollisions(rayNP):
					queue = CollisionHandlerQueue()
					self.traverser.addCollider(rayNP, queue)
					self.traverser.traverse(self.graphRoot)
					self.traverser.clearColliders()
					queue.sortEntries()
					return queue
				
				queue = getCollisions(pickerNP)
				for i in range(queue.getNumEntries()):
					hit = queue.getEntry(i)
					id = hit.getIntoNodePath().getParent().getName()
					if len(id) > 0:
						node = self.getNode(id)
						if node != None:
							if node != self.cameraFocusNode or not node.expanded:
								node.node.setScale(2.0 * node.defaultScale)
							self.lastHighlightedNode = node
							break
		
		cameraFocus = Point3() if self.cameraFocusNode == None else self.cameraFocusNode.node.getPos()
		cameraOrbitRadius = 50 + (0 if self.cameraFocusNode == None else self.cameraFocusNode.childOrbitRadius * 2)
		if self.cameraFocusInterval < 1.0:
			lastCameraFocus = Point3() if self.lastCameraFocusNode == None else self.lastCameraFocusNode.node.getPos()
			lastCameraOrbitRadius = 50 + (0 if self.lastCameraFocusNode == None else self.lastCameraFocusNode.childOrbitRadius * 2)
			cameraFocus = (cameraFocus * self.cameraFocusInterval) + (lastCameraFocus * (1.0 - self.cameraFocusInterval))
			cameraOrbitRadius = (cameraOrbitRadius * self.cameraFocusInterval) + (lastCameraOrbitRadius * (1.0 - self.cameraFocusInterval))
			self.cameraFocusInterval = min(1.0, self.cameraFocusInterval + dt)
		base.camera.setPos(cameraFocus + (Vec3(math.cos(self.cameraYaw) * math.sin(self.cameraPitch), math.sin(self.cameraYaw) * math.sin(self.cameraPitch), math.cos(self.cameraPitch)) * cameraOrbitRadius))
		base.camera.lookAt(Point3(cameraFocus))
		
		self.space.autoCollide()
		self.world.quickStep(dt)
		self.contactGroup.empty() # Clear the contact joints
		for node in self.nodes:
			node.update(dt)
	
	def draw(self):
		self.drawer.begin(base.cam, render)
		for a, b in self.segments:
			self.drawer.segment(Vec3(a.node.getPos()), Vec3(b.node.getPos()), Vec4(0, 0, 1, 1), .05, Vec4(1, 1, 1, 1))
		self.drawer.end()
	
	def destroy(self):
		for node in self.nodes:
			node.destroy()
		self.world.destroy()
		self.space.destroy()

class MeshGraphNode:
	def __init__(self, graphRoot, parent, orbitRadius, id, text, data, world, space):
		if id == None:
			self.id = uuid4().hex
		else:
			self.id = id
		self.text = text.encode("ascii", "replace")
		self.text = boldStartRegex.sub("\1strong\1", self.text)
		self.text = boldEndRegex.sub("\2", self.text)
		self.text = italicStartRegex.sub("\1em\1", self.text)
		self.text = italicEndRegex.sub("\2", self.text)
		self.text = subscriptStartRegex.sub("\1sub\1", self.text)
		self.text = subscriptEndRegex.sub("\2", self.text)
		self.text = superscriptStartRegex.sub("\1sup\1", self.text)
		self.text = superscriptEndRegex.sub("\2", self.text)
		self.text = underlineStartRegex.sub("\1u\1", self.text)
		self.text = underlineEndRegex.sub("\2", self.text)
		self.text = unescape(allTagsRegex.sub("", self.text))
		self.data = data
		self.parent = parent
		self.orbitRadius = orbitRadius
		self.physicsRadius = 3
		self.randomX = (random() * 2.0) - 1.0
		self.randomY = (random() * 2.0) - 1.0
		self.randomZ = (random() * 2.0) - 1.0
		self.expanded = False
		self.hasImage = False
		
		self.childOrbitRadius = 25
		
		self.defaultScale = 1.0
		
		t = TextNode("")
		t.setFont(loader.loadFont("Font.ttf"))
		t.setWordwrap(10.0)
		t.setAlign(TextNode.ACenter)
		self.node = graphRoot.attachNewNode(t)
		
		self.recalculateOffset()
		
		self.node.setName(self.id)
		self.node.setTwoSided(True)
		self.node.setBillboardPointEye()
		self.constructCollisionNode()
		
		self.links = {}
		self.body = OdeBody(world)
		M = OdeMass()
		M.setSphere(0.05, self.physicsRadius)
		self.body.setMass(M)
		self.geometry = OdeSphereGeom(space, self.physicsRadius)
		self.geometry.setBody(self.body)
		space.setSurfaceType(self.geometry, 0)
		self.body.setPosition((self.parent.body.getPosition() if self.parent != None else Vec3(0, 0, 0)) + self.offset)
		self.node.setPos(self.body.getPosition())
	
	def constructCollisionNode(self):
		self.collisionNode = CollisionNode("cnode")
		self.collisionNode.addSolid(CollisionSphere(0, 0, 0, self.physicsRadius))
		self.collisionNodePath = self.node.attachNewNode(self.collisionNode)
	
	def setImage(self, source):
		self.hasImage = True
		p = PNMImage()
		p.read(StringStream(source))
		tex = Texture()
		tex.load(p)
		
		cm = CardMaker("")
		scale = 0.1
		cm.setFrame(-tex.getXSize() * 0.5 * scale, tex.getXSize() * 0.5 * scale, -tex.getYSize() * 0.5 * scale, tex.getYSize() * 0.5 * scale)
		cm.setHasUvs(True)
		
		newNode = self.node.getParent().attachNewNode(cm.generate())
		newNode.setTexture(tex)
		newNode.setTwoSided(True)
		newNode.setBillboardPointEye()
		newNode.setPos(self.node.getPos())
		newNode.setName(self.id)
		self.collisionNodePath.reparentTo(newNode)
		self.node.removeNode()
		self.node = newNode
	
	def recalculateOffset(self, selected = False):
		if self.hasImage:
			self.defaultScale = 2.0 if selected else 1.0
			self.node.setScale(self.defaultScale)
		else:
			t = self.node.node()
			if selected:
				t.setWordwrap(25.0)
				t.setText(self.text)
			else:
				t.setWordwrap(10.0)
				t.setText(self.text[0:35] + ("..." if len(self.text) > 35 else ""))
		
		radius = self.orbitRadius + (30 if selected else 0)
		if self.parent != None:
			if self.parent.offset.length() == 0.0:
				offset = Vec3(self.randomX, self.randomY, self.randomZ)
				offset.normalize()
				self.offset = offset * radius
			else:
				parentOffset = Vec3(self.parent.offset)
				parentOffset.normalize()
				right = parentOffset.cross(Vec3.up())
				up = parentOffset.cross(right)
				offset = parentOffset + (right * self.randomX) + (up * self.randomY)
				offset.normalize()
				self.offset = offset * radius
		else:
			self.offset = Vec3()
	
	def update(self, dt):
		if self.parent == None:
			self.body.setPosition(self.node.getPos())
		else:
			bodyPosition = self.body.getPosition()
			force = ((self.parent.body.getPosition() + self.offset) - bodyPosition) * 800 * dt
			force -= self.body.getLinearVel() * 800 * dt
			self.body.addForce(force)
			self.node.setPos(bodyPosition)
	
	def link(self, key, child):
		self.links[key] = child
	
	def destroy(self):
		self.geometry.destroy()
		self.body.destroy()
		self.node.removeNode()